@@ -4,10 +4,12 @@ import logging
 import struct
 import sys
 from collections import namedtuple
+from copy import copy
 import numpy as np
 
 
 log = logging.getLogger(__name__)
+logging.basicConfig()
 log.setLevel(logging.WARNING)
 # To adjust the log level for this module from a script, use eg:
 # logging.getLogger(tdms.__name__).setLevel(logging.DEBUG)
@@ -113,6 +115,7 @@ class TdmsFile(object):
                 self._read_segments(tdms_file)
 
     def _read_segments(self, tdms_file):
+        # Read metadata first to work out how much space we need
         previous_segment = None
         while True:
             try:
@@ -122,37 +125,54 @@ class TdmsFile(object):
                 break
             segment.read_metadata(tdms_file, self.objects,
                     previous_segment)
-            segment.read_raw_data(tdms_file, self.objects)
 
             self.segments.append(segment)
             previous_segment = segment
             if segment.next_segment_pos is None:
                 break
             else:
-                if(tdms_file.tell() != segment.next_segment_pos):
-                    log.warning("Did not read to the end of the "
-                            "segment, there may be unread data.")
-                    tdms_file.seek(segment.next_segment_pos)
+                tdms_file.seek(segment.next_segment_pos)
 
-    def _path(self, group=None, channel=None):
-        path = '/'
-        if group is not None:
-            path += "'" + group.replace("'", "''") + "'"
-        if channel is not None:
-            path += "/'" + channel.replace("'", "''") + "'"
-        return path
+        # Allocate space for data
+        for object in self.objects.values():
+            object._initialise_data()
 
-    def object(self, group, channel=None):
-        """Get a TDMS Object from the file
+        # Now actually read all the data
+        for segment in self.segments:
+            segment.read_raw_data(tdms_file)
 
-        :param group: Name of the group of the object.
-        :keyword channel: The name of the channel, or None to get the
-            group object.
+    def _path(self, *args):
+        """Convert group and channel to object path"""
+
+        return ('/' + '/'.join(
+                ["'" + arg.replace("'", "''") + "'" for arg in args]))
+
+    def object(self, *path):
+        """Get a TDMS object from the file
+
+        :param path: The object group and channel. Providing no channel
+            returns a group object, and providing no channel or group
+            will return the root object.
         :rtype: :class:`TdmsObject`
 
+        For example, to get the root object::
+
+            object()
+
+        To get a group::
+
+            object("group_name")
+
+        To get a channel::
+
+            object("group_name", "channel_name")
         """
 
-        return self.objects[self._path(group, channel)]
+        object_path = self._path(*path)
+        try:
+            return self.objects[object_path]
+        except KeyError:
+            raise KeyError("Invalid object path: %s" % object_path)
 
     def groups(self):
         """Return the names of groups in the file
@@ -194,6 +214,7 @@ class _TdmsSegment(object):
         """Read the lead in section of a segment"""
 
         self.position = f.tell()
+        self.num_chunks = 0
         self.ordered_objects = []
 
         # First four bytes should be TDSm
@@ -229,8 +250,9 @@ class _TdmsSegment(object):
         (self.next_segment_offset, self.raw_data_offset) = (
                 struct.unpack('<QQ', s))
 
-        # Calculate next segment position
+        # Calculate data and next segment position
         lead_size = 7 * 4
+        self.data_position = self.position + lead_size + self.raw_data_offset
         if self.next_segment_offset == struct.unpack('<Q', b'\xFF' * 8)[0]:
             # This can happen if Labview crashes
             log.warning("Last segment of file has unknown size")
@@ -246,10 +268,19 @@ class _TdmsSegment(object):
         """Read segment metadata section and update object information"""
 
         if not self.toc["kTocMetaData"]:
-            self.ordered_objects = previous_segment.ordered_objects
+            try:
+                self.ordered_objects = previous_segment.ordered_objects
+            except AttributeError:
+                raise ValueError("kTocMetaData is set for segment but "
+                        "there is no previous segment")
+            self.calculate_chunks()
             return
         if not self.toc["kTocNewObjList"]:
-            self.ordered_objects = previous_segment.ordered_objects
+            # In this case, there can be a list of new objects that
+            # are appended, or previous objects can also be repeated
+            # if their properties change
+            self.ordered_objects = [
+                    copy(o) for o in previous_segment.ordered_objects]
 
         log.debug("Reading metadata at %d" % f.tell())
 
@@ -268,20 +299,44 @@ class _TdmsSegment(object):
             else:
                 obj = TdmsObject(object_path)
                 objects[object_path] = obj
-            obj._read_metadata(f)
-            if (self.toc["kTocNewObjList"] or
-                    object_path not in [o.path for o in self.ordered_objects]):
-                self.ordered_objects.append(obj)
 
-    def read_raw_data(self, f, objects):
-        """Read signal data from file"""
+            # Update or create list of ordered segment objects, re-using
+            # any properties from previous segments
+            if self.toc["kTocNewObjList"]:
+                if obj._previous_segment_object is not None:
+                    segment_obj = copy(obj._previous_segment_object)
+                else:
+                    segment_obj = _TdmsSegmentObject(obj)
+                segment_obj._read_metadata(f)
+                self.ordered_objects.append(segment_obj)
+            else:
+                obj_index = [
+                        i for i, o in enumerate(self.ordered_objects)
+                        if o.tdms_object is obj]
+                if len(obj_index) == 0:
+                    segment_obj = _TdmsSegmentObject(obj)
+                    segment_obj._read_metadata(f)
+                    self.ordered_objects.append(segment_obj)
+                else:
+                    obj_index = obj_index[0]
+                    segment_obj = self.ordered_objects[obj_index]
+                    # Here the data type and number of data values is
+                    # always set, but properties might be updated
+                    segment_obj._read_metadata(f)
+            obj._previous_segment_object = segment_obj
 
-        if not self.toc["kTocRawData"]:
-            return
+        self.calculate_chunks()
+
+    def calculate_chunks(self):
+        """
+        Work out the number of chunks the data is in, for cases
+        where the meta data doesn't change at all so there is no
+        lead in.
+
+        Also increments the number of values for objects in this
+        segment, based on the number of chunks.
+        """
 
-        # Work out the number of chunks the data is in, for cases
-        # where the meta data doesn't change at all so there is no
-        # lead in.
         data_size = sum([
                 o.data_size
                 for o in self.ordered_objects])
@@ -293,21 +348,37 @@ class _TdmsSegment(object):
             if total_data_size != data_size:
                 raise ValueError("Zero channel data size but non-zero data "
                         "length based on segment offset.")
+            self.num_chunks = 0
             return
         if total_data_size % data_size != 0:
             raise ValueError("Data size %d is not a multiple of the "
                     "chunk size %d" % (total_data_size, data_size))
         else:
-            num_chunks = total_data_size // data_size
+            self.num_chunks = total_data_size // data_size
+
+        # Update data count for object
+        for obj in self.ordered_objects:
+            obj.tdms_object.number_values += (
+                    obj.number_values * self.num_chunks)
+
+    def read_raw_data(self, f):
+        """Read signal data from file"""
+
+        if not self.toc["kTocRawData"]:
+            return
+
+        f.seek(self.data_position)
+
+        total_data_size = self.next_segment_offset - self.raw_data_offset
         log.debug("Reading %d bytes of data at %d in %d chunks" %
-                (total_data_size, f.tell(), num_chunks))
+                (total_data_size, f.tell(), self.num_chunks))
 
         if self.toc['kTocBigEndian']:
             endianness = '>'
         else:
             endianness = '<'
 
-        for chunk in range(num_chunks):
+        for chunk in range(self.num_chunks):
             if self.toc["kTocInterleavedData"]:
                 log.debug("Data is interleaved")
                 data_objects = [o for o in self.ordered_objects if o.has_data]
@@ -332,7 +403,7 @@ class _TdmsSegment(object):
 
                 for obj in self.ordered_objects:
                     if obj.has_data:
-                        obj._update_data(object_data[obj.path])
+                        obj.tdms_object._update_data(object_data[obj.path])
 
     def _read_interleaved_numpy(self, f, data_objects, endianness):
         """Read interleaved data where all channels have a numpy type"""
@@ -357,7 +428,7 @@ class _TdmsSegment(object):
             # be correct
             object_data.dtype = (
                     np.dtype(obj.data_type.nptype).newbyteorder(endianness))
-            obj._update_data(object_data)
+            obj.tdms_object._update_data(object_data)
             data_pos += obj.data_type.length
 
     def _read_interleaved(self, f, data_objects, endianness):
@@ -377,7 +448,7 @@ class _TdmsSegment(object):
                             obj._read_value(f, endianness))
                     points_added[obj.path] += 1
         for obj in data_objects:
-            obj._update_data(object_data[obj.path])
+            obj.tdms_object._update_data(object_data[obj.path])
 
 
 class TdmsObject(object):
@@ -396,16 +467,93 @@ class TdmsObject(object):
         self.path = path
         self.data = None
         self.properties = {}
-        self.raw_data_index = 0
         self.data_type = None
         self.dimension = 1
         self.number_values = 0
-        self.data_size = 0
-        self.has_data = True
+        self.has_data = False
+        self._previous_segment_object = None
 
     def __repr__(self):
         return "<TdmsObject with path %s>" % self.path
 
+    def property(self, property_name):
+        """Returns the value of a TDMS property
+
+        :param property_name: The name of the property to get.
+        :returns: The value of the requested property.
+        :raises: KeyError if the property isn't found.
+
+        """
+
+        return self.properties[property_name]
+
+    def time_track(self):
+        """Return an array of time for this channel
+
+        This depends on the object having the wf_increment
+        and wf_start_offset properties defined.
+
+        :rtype: NumPy array.
+        :raises: KeyError if required properties aren't found
+
+        """
+
+        try:
+            increment = self.property('wf_increment')
+            offset = self.property('wf_start_offset')
+        except KeyError:
+            raise KeyError("Object does not have time properties available.")
+
+        return np.linspace(
+                offset,
+                offset + (len(self.data) - 1) * increment,
+                len(self.data))
+
+    def _initialise_data(self):
+        """Initialise data array to zeros"""
+
+        if self.number_values > 0:
+            if self.data_type.nptype is None:
+                self.data = []
+            else:
+                self.data = np.zeros(
+                        self.number_values, dtype=self.data_type.nptype)
+                self._data_insert_position = 0
+
+    def _update_data(self, new_data):
+        """Update the object data with a new array of data"""
+
+        log.debug("Adding %d data points to data for %s" %
+                (len(new_data), self.path))
+        if self.data is None:
+            self.data = new_data
+        else:
+            if self.data_type.nptype is not None:
+                data_pos = (self._data_insert_position,
+                        self._data_insert_position + len(new_data))
+                self._data_insert_position += len(new_data)
+                self.data[data_pos[0]:data_pos[1]] = new_data
+            else:
+                self.data.extend(new_data)
+
+
+class _TdmsSegmentObject(object):
+    """
+    Describes an object in an individual TDMS file segment
+    """
+
+    def __init__(self, tdms_object):
+        self.tdms_object = tdms_object
+
+        self.path = tdms_object.path
+        self.raw_data_index = 0
+        self.number_values = 0
+        self.data_size = 0
+        self.has_data = True
+        self.data_type = None
+        self.dimension = 1
+        self.properties = {}
+
     def _read_metadata(self, f):
         """Read object metadata and update object information"""
 
@@ -422,11 +570,17 @@ class TdmsObject(object):
             pass
         else:
             self.has_data = True
-            index_length = self.raw_data_index
+            self.tdms_object.has_data = True
 
             # Read the data type
             s = f.read(4)
             self.data_type = tdsDataTypes[struct.unpack("<L", s)[0]]
+            if (self.tdms_object.data_type is not None and
+                    self.data_type != self.tdms_object.data_type):
+                raise ValueError("Segment object doesn't have the same data "
+                        "type as previous segments.")
+            else:
+                self.tdms_object.data_type = self.data_type
 
             # Read data dimension
             s = f.read(4)
@@ -465,47 +619,7 @@ class TdmsObject(object):
                 value = read_type(f, prop_data_type, '<')
             self.properties[prop_name] = value
             log.debug("Property %s: %s" % (prop_name, value))
-
-    def property(self, property_name):
-        """Returns the value of a TDMS property
-
-        :param property_name: The name of the property to get.
-        :returns: The value of the requested property.
-        :raises: KeyError if the property isn't found.
-
-        """
-
-        return self.properties[property_name]
-
-    def time_track(self):
-        """Return an array of time for this channel
-
-        This depends on the object having the wf_increment
-        and wf_start_offset properties defined.
-
-        :rtype: NumPy array.
-        :raises: KeyError if required properties aren't found
-
-        """
-
-        try:
-            increment = self.property('wf_increment')
-            offset = self.property('wf_start_offset')
-        except KeyError:
-            raise KeyError("Object does not have time properties available.")
-
-        return np.linspace(
-                offset,
-                offset + (len(self.data) - 1) * increment,
-                len(self.data))
-
-    def _new_segment_data(self):
-        """Return a new array to read the data of the current section into"""
-
-        if self.data_type.nptype is not None:
-            return np.zeros(self.number_values, dtype=self.data_type.nptype)
-        else:
-            return [None] * self.number_values
+        self.tdms_object.properties.update(self.properties)
 
     def _read_value(self, file, endianness):
         """Read a single value from the given file"""
@@ -528,15 +642,10 @@ class TdmsObject(object):
             data[i] = read_type(file, self.data_type, endianness)
         return data
 
-    def _update_data(self, new_data):
-        """Update the object data with a new array of data"""
+    def _new_segment_data(self):
+        """Return a new array to read the data of the current section into"""
 
-        log.debug("Adding %d data points to data for %s" %
-                (len(new_data), self.path))
-        if self.data is None:
-            self.data = new_data
+        if self.data_type.nptype is not None:
+            return np.zeros(self.number_values, dtype=self.data_type.nptype)
         else:
-            if self.data_type.nptype is not None:
-                self.data = np.append(self.data, new_data)
-            else:
-                self.data.extend(new_data)
+            return [None] * self.number_values
